//-------------------------------------------------------------------------
//   Copyright 2002-2018 National Technology & Engineering Solutions of
//   Sandia, LLC (NTESS).  Under the terms of Contract DE-NA0003525 with
//   NTESS, the U.S. Government retains certain rights in this software.
//
//   This file is part of the Xyce(TM) Parallel Electrical Simulator.
//
//   Xyce(TM) is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   Xyce(TM) is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with Xyce(TM).
//   If not, see <http://www.gnu.org/licenses/>.
//-------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Purpose       : This file contains the functions to manage fourier objects
// Special Notes :
// Creator       : Heidi Thornquist, SNL, Electrical and Microsystem Modeling
// Creation Date : 03/10/2009
//
//-----------------------------------------------------------------------------

#include <Xyce_config.h>

#include <iostream>

#include <N_DEV_DeviceMgr.h>
#include <N_IO_FourierMgr.h>
#include <N_IO_CircuitBlock.h>
#include <N_IO_NetlistImportTool.h>
#include <N_IO_OptionBlock.h>
#include <N_IO_PkgOptionsMgr.h>
#include <N_IO_SpiceSeparatedFieldTool.h>

#include <N_ERH_ErrorMgr.h>
#include <N_UTL_ExtendedString.h>
#include <N_UTL_FeatureTest.h>
#include <N_UTL_Math.h>
#include <N_UTL_Op.h>
#include <N_UTL_OpBuilder.h>
#include <N_UTL_OptionBlock.h>

#include <Teuchos_ScalarTraits.hpp>

namespace Xyce {
namespace IO {
//-----------------------------------------------------------------------------
// Function      : FourierMgr::FourierMgr
// Purpose       : constructor
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, SNL, Electrical and Microsystem Modeling
// Creation Date : 03/10/2009
//-----------------------------------------------------------------------------
FourierMgr::FourierMgr(
  const std::string &   netlist_filename)
  : netlistFilename_(netlist_filename),
    numFreq_(10),
    gridSize_(200),
    calculated_(false)
{
  // Initialize first output variable pointer.  This vector is used to determine
  // which output variables are associated with which fundamental frequency.
  outputVarsPtr_.push_back( 0 );
}

//-----------------------------------------------------------------------------
// Function      : FourierMgr::FourierMgr
// Purpose       : destructor
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, SNL, Electrical and Microsystem Modeling
// Creation Date : 03/10/2009
//-----------------------------------------------------------------------------
FourierMgr::~FourierMgr()
{
  for (Util::Op::OpList::iterator it = outputVars_.begin(); it != outputVars_.end(); ++it)
    delete *it;

}

//-----------------------------------------------------------------------------
// Function      : FourierMgr::fixupFourierParameters
// Purpose       : This function extracts info from the maps (freqSolVarMap_,
//                 freqNamesMap_ and freqNumOutputVarsMap_) created in 
//                 addFourierAnalysis(), in order to support multiple .FOUR
//                 lines that may have different fundamental frequencies. It
//                 creates the vectors (names_, freqVector_ and depSolVarIterVector_) 
//                 used by the existing code.  Simplifying this function likely 
//                 involves re-writing the rest of the .FOUR code.
// Special Notes : 
// Scope         : public
// Creator       : Not Sure (but likely Dave Baur)
// Creation Date : Not Sure
//-----------------------------------------------------------------------------
void FourierMgr::fixupFourierParameters(Parallel::Machine comm, const Util::Op::BuilderManager &op_builder_manager) 
{
  // this copy (from multi-map to vector) is done in order to support multiple .FOUR lines
  // The multi-map allows us to keep a sorted list of which solution variables are associated with
  // which fundamental frequency.
  for(std::multimap<double,Util::Param>::iterator it = freqSolVarMap_.begin(); it != freqSolVarMap_.end(); ++it ) 
  {
    // this vector is used by the makeOps function call below
    depSolVarIterVector_.push_back( it->second );   
  }

  Util::Op::makeOps(comm, op_builder_manager, NetlistLocation(), depSolVarIterVector_.begin(), depSolVarIterVector_.end(), std::back_inserter(outputVars_));

  // These copies (from set (or map) to vector) were done so that the uses of freqVector_ and names_
  // in the rest of the calculations and output functions did not have to be changed.  The 
  // freqNumOutputVarsMap_ contains the set of fundamental frequencies, and the number of output
  // variables at each frequency.  
  int ovPtrPos=0;
  for (std::map<double,int>::iterator it=freqNumOutputVarsMap_.begin(); it!=freqNumOutputVarsMap_.end(); ++it, ++ovPtrPos)
  {
    freqVector_.push_back(it->first);
    outputVarsPtr_.push_back( outputVarsPtr_[ovPtrPos] + it->second );
  }

  // The freqNamesMap_ has a listing of "names" for the printResult_() function. 
  // It is needed in case the user intermixes .FOUR lines such as below. 
  //     .FOUR 1 V(1)
  //     .FOUR 2 V(2)
  //     .FOUR 1 V(3)
  for (std::multimap<double,std::string>::iterator it=freqNamesMap_.begin(); it!=freqNamesMap_.end(); ++it)
  {
    names_.push_back(it->second);
  } 

  // some error checking that the vectors have the right sizes and values
  if ( (names_.size() != outputVars_.size()) || ( outputVarsPtr_[outputVarsPtr_.size()-1] != outputVars_.size()) )
  {
    Report::UserFatal0() << "Unknown error while parsing .FOUR lines";
  }
}

//-----------------------------------------------------------------------------
// Function      : FourierMgr::addFourierAnalysis
// Purpose       : Entry point when .four lines are pased in the netlist
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, SNL, Electrical and Microsystem Modeling
// Creation Date : 03/10/2009
//-----------------------------------------------------------------------------
bool FourierMgr::addFourierAnalysis(const Util::OptionBlock & fourierBlock)
{
  // based on what's in the option block passed in, we
  // create the needed fourier instance
  double freq;
  Util::ParamList variableList;   // Used to help register lead current requests with device manager.

  if (DEBUG_IO)
    Xyce::dout() << "In FourierMgr::addFourier" << std::endl
                 << ".FOUR line passed was: " << std::endl << fourierBlock << std::endl;

  const Util::Param *param = Util::findParameter(fourierBlock.begin(), fourierBlock.end(), "FREQ");
  if (!param)
  {
    // this shouldn't happen, but catch it if does
    Report::DevelFatal0() << "Missing FREQ in .FOUR line!";
  }

  for (Util::ParamList::const_iterator currentParamIt = fourierBlock.begin(); currentParamIt != const_cast<Util::OptionBlock &>(fourierBlock).end(); ++currentParamIt)
  {
    std::string name = "";
    std::string tag = currentParamIt->tag();

    if( tag == "FREQ" )
    {
      freq = currentParamIt->getImmutableValue<double>();
      if ( freq < 0 )
      {
        // a negative value for freq will cause a core dump later in the code.
        Report::DevelFatal0() << "Illegal FREQ value in .FOUR line!";
      }
      else
      {
        // a map is used here, in order to support multiple .FOUR lines with the same
        // fundamental frequency.  It stores a sorted version of the values of how many 
        // output variables are requested at each fundamental frequency
        freqNumOutputVarsMap_.insert(std::make_pair(freq,0));
      }
    }
    else if( (tag == "V") || (tag[0] == 'I') || (tag == "P") || (tag == "W") || (tag == "N") )
    {
      // tag[0] is used for I because branch currents for transistors can have two
      // characters.  An example is IS for the M-Device.
      int nodes = currentParamIt->getImmutableValue<int>();
      Util::Param aParam;
      aParam.set( tag, nodes );
      name += tag;
      name += "(";

      // add to list of variables.  This will be used later in netlist parsing 
      // to enable lead currents in the device manager.
      variableList.push_back(aParam);

      // here we just store the needed parts of V(a) or v(a,b) or I(device).
      // only the v(a,b) case will need an extra node in the outputVars_ array.
      // use list::insert() to keep an iterator pointing to this spot
    
      if ( freqNumOutputVarsMap_.size() == 0 )
      {
        // expected that a freq would have been found at this point, since the FREQ value 
        // comes before the ov variables on a .FOUR line.
        Report::DevelFatal0() << "Unable to parse .FOUR line";
      }
      else
      {
        // map is used here, in order to support multiple .FOUR lines, potentially with
        // different fundamental frequencies.
        freqSolVarMap_.insert(std::make_pair(freq,aParam));
        ++freqNumOutputVarsMap_[freq];
        for( int i=0; i<nodes; i++ )
        {
          currentParamIt++;
          aParam.set( currentParamIt->tag(), currentParamIt->getImmutableValue<double>() );
          freqSolVarMap_.insert(std::make_pair(freq,aParam));
          name += currentParamIt->tag(); 
          if (i != nodes-1 && nodes > 1) name += ",";     

          // add to list of variables.  This will be used later in netlist parsing 
          // to enable lead currents in the device manager.
          variableList.push_back(aParam);
        }
        name += ")";
      }
    }
    else if( Xyce::Util::hasExpressionTag(tag) )
    {
      if ( freqNumOutputVarsMap_.size() == 0 )
      {
        // expected that a freq would have been found at this point, since the FREQ value 
        // comes before the ov variables on a .FOUR line.
        Report::DevelFatal0() << "Unable to parse .FOUR line";
      }
      else
      {
        freqSolVarMap_.insert(std::make_pair(freq,*currentParamIt));
        ++freqNumOutputVarsMap_[freq];
        name = tag;

        // add to list of variables.  This will be used later in netlist parsing 
        // to enable lead currents in the device manager.
        Util::Param aParam;
        aParam.set( currentParamIt->tag(), currentParamIt->tag() );
        variableList.push_back(aParam);
      }
    }

    // Save voltage or current variable name.
    if (name != "")
    {
      // a map is used here, in order to support multiple .FOUR lines, potentially with
      // different fundamental frequencies.
      freqNamesMap_.insert(std::make_pair(freq,name));
    }
  }
  
  // Used to help register lead current requests with device manager.
  getLeadCurrentDevices(variableList, devicesNeedingLeadCurrents_);

  return true;
}

//-----------------------------------------------------------------------------
// Function      : FourierMgr::updateFourierData
// Purpose       : Called during the simulation to update the fourier objects
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, SNL, Electrical and Microsystem Modeling
// Creation Date : 03/10/2009
//-----------------------------------------------------------------------------
void FourierMgr::updateFourierData(Parallel::Machine comm, const double circuitTime, const Linear::Vector *solnVec, 
  const Linear::Vector *stateVec, const Linear::Vector * storeVec,
  const Linear::Vector *lead_current_vector, const Linear::Vector *junction_voltage_vector, 
  const Linear::Vector *lead_current_dqdt_vector)
{
  // Save the time.
  time_.push_back(circuitTime);

  //for (unsigned int i=0; i<outputVars_.size(); ++i)
  int vecIndex = 0;
  for (std::vector<Util::Op::Operator *>::const_iterator it = outputVars_.begin(); it != outputVars_.end(); ++it)
  {
    double retVal = getValue(comm, *(*it), Util::Op::OpData(vecIndex, solnVec, 0, stateVec, storeVec, 0, lead_current_vector, 0, junction_voltage_vector)).real();
    outputVarsValues_.push_back(retVal);
    vecIndex++;
  }
}


//-----------------------------------------------------------------------------
// Function      : FourierMgr::outputResults
// Purpose       : Output fourier results at end of simulation
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, SNL, Electrical and Microsystem Modeling
// Creation Date : 03/10/2009
//-----------------------------------------------------------------------------
void FourierMgr::outputResults(std::ostream& outputStream)
{
  // Only calculate something if a .four line was encountered and transient data was collected.
  int numOutVars = outputVars_.size();

  if ( numOutVars && !time_.empty() && !calculated_ )
  {
    // Calculated the fourier coefficients for the given nodes.
    getLastPeriod_();
    interpolateData_();
    calculateFT_();
    calculated_ = true;
  }

  // Output the information to the outputStream 
  printResult_( outputStream );

}


//-----------------------------------------------------------------------------
// Function      : FourierMgr::getLastPeriod_()
// Purpose       : finds the indices to access the last period of simulation
// Special Notes :
// Scope         : private
// Creator       : Heidi Thornquist, SNL, Electrical Models & Simulation
// Creation Date : 6/21/13
//-----------------------------------------------------------------------------
void FourierMgr::getLastPeriod_()
{
  // We want to do the analysis on only the last period of the transient window. So here we find the indices
  // to access the endpoints of that interval.
  int numPoints = time_.size();
  int prdEnd = numPoints - 1;
  double endTime = time_[prdEnd];

  int nFreq = freqVector_.size();
  lastPrdStart_.resize( nFreq );
  prdStart_.resize( nFreq );

  for (int i=0; i<nFreq; ++i)
  {
    // Use this formulation to compute lastPrdStart to avoid issues with 32-bit arithmetic.
    lastPrdStart_[i] = (freqVector_[i]*endTime - 1.0)/freqVector_[i];

    if ( Teuchos::ScalarTraits<double>::magnitude( lastPrdStart_[i] ) < 
         Teuchos::ScalarTraits<double>::eps() )
    {
      lastPrdStart_[i] = 0.0;
      prdStart_[i] = 0;
    } 
    else if (lastPrdStart_[i] > 0)
    {
      // Initialize prdStart_ to be the index of the last element in time_.
      // Then scan until time_[i] <= endTime_ - period_.
      prdStart_[i] = prdEnd;
      while (time_[prdStart_[i]] > lastPrdStart_[i])
      {
        prdStart_[i] -= 1;
      }
    }
    else
    {
      std::string msg = "Error: The period is greater than the length of the time simulation. Exiting.";
      Report::UserFatal() << msg;
    }
  }
} 

//-----------------------------------------------------------------------------
// Function      : FourierMgr::interpolateData_()
// Purpose       : evaluates interpolating polynomial at equidistant time pts
// Special Notes :
// Scope         : private
// Creator       : Heidi Thornquist, SNL, Electrical Models & Simulation
// Creation Date : 6/21/13
//-----------------------------------------------------------------------------
bool FourierMgr::interpolateData_()
{
  double A, B, C;

  int numOutVars = outputVars_.size();
  int numPoints = time_.size();
  int nFreq = freqVector_.size();

  newTime_.resize(nFreq * gridSize_,0.0);
  newValues_.resize(numOutVars * gridSize_,0.0);

  for (int j=0; j<nFreq; ++j)
  {
    // Get the number of solution variables associated with this frequency
    int numCurrVars = outputVarsPtr_[j+1] - outputVarsPtr_[j];

    // Get the number of time points in the final period
    int nData = numPoints-prdStart_[j];
    double period = 1.0/freqVector_[j];
    std::vector<double> h(nData-1, 0.0);
    std::vector<double> b(nData-1, 0.0);
    std::vector<double> u(nData-1, 0.0);
    std::vector<double> v(nData-1, 0.0);
    std::vector<double> z(nData, 0.0);

    // Compute new, equally spaced time points.
    double step = period/gridSize_;

    newTime_[gridSize_*j] = lastPrdStart_[j];
    for (int i = 1; i < gridSize_; i++) 
    {
      newTime_[gridSize_*j + i] = newTime_[gridSize_*j + i-1] + step;
    }
  
    // Loop over all the data from the output variables associated with this frequency 
    for (int k=0; k<numCurrVars; k++)
    { 
      int offset = outputVarsPtr_[j] + k;

      // Cubic spline interpolation. We first need to find the z's.
      for (int i = 0; i < nData-1; i++)
      {
        h[i] = time_[i+1+prdStart_[j]]-time_[i+prdStart_[j]];
        b[i] = (6/h[i])*(outputVarsValues_[(i+1+prdStart_[j])*numOutVars + offset]
                         - outputVarsValues_[(i+prdStart_[j])*numOutVars + offset]);
      }

      u[1] = 2*(h[0]+h[1]);
      v[1] = b[1]-b[0];

      for (int i=2; i < nData-1; i++)
      {
        u[i] = 2*(h[i]+h[i-1])-((h[i-1])*(h[i-1]))/u[i-1];
        v[i] = b[i]-b[i-1]-(h[i-1]*v[i-1])/u[i-1];
      }

      z[nData-1] = 0;
      for (int i=nData-2; i > 0; i--)
      {
        z[i] = (v[i]-h[i]*z[i+1])/u[i];
      }
      z[0] = 0;

      // Calculate the new values at the new time points.
      for (int i = 0; i < gridSize_; i++)
      { 
        int idx = nData-1;
        while ((newTime_[gridSize_*j + i]-time_[idx+prdStart_[j]]) < 0)
        {
          idx--;
        }
        A = (z[idx+1]-z[idx])/(6*h[idx]);
        B = z[idx]/2;
        C = -(h[idx]/6)*z[idx+1]-(h[idx]/3)*z[idx]+(outputVarsValues_[(idx+1+prdStart_[j])*numOutVars + offset]
                                                    - outputVarsValues_[(idx+prdStart_[j])*numOutVars + offset])/h[idx];
        newValues_[gridSize_*offset + i] = outputVarsValues_[(idx+prdStart_[j])*numOutVars + offset] + (newTime_[gridSize_*j + i]-time_[idx+prdStart_[j]])*(C+(newTime_[gridSize_*j + i]-time_[idx+prdStart_[j]])*(B+(newTime_[gridSize_*j + i]-time_[idx+prdStart_[j]])*A));
      }
    } 
  }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : FourierMgr::calculateFT_()
// Purpose       : performs fourier analysis
// Special Notes :
// Scope         : private
// Creator       : Heidi Thornquist, SNL, Electrical Models & Simulation
// Creation Date : 6/21/13
//-----------------------------------------------------------------------------
void FourierMgr::calculateFT_()
{ 
  int numOutVars = outputVars_.size();
  int nFreq = freqVector_.size();

  mag_.resize(numFreq_*numOutVars, 0.0);
  phase_.resize(numFreq_*numOutVars, 0.0);
  nmag_.resize(numFreq_*numOutVars, 0.0);
  nphase_.resize(numFreq_*numOutVars, 0.0);
  freq_.resize(numFreq_*nFreq, 0.0);
  thd_.resize(numOutVars, 0.0);

  // Compute frequencies 
  for (int j=0; j<nFreq; ++j)
  {
    for (int i=0; i < numFreq_; i++)
    {
      freq_[j*numFreq_ + i] = i * freqVector_[j];
    }
  }
   
  // Perform Fourier analysis for all the output variables. 
  for (int k=0; k < numOutVars; k++)
  { 
    for (int i=0; i < gridSize_; i++)
    { 
      for (int j=0; j < numFreq_; j++)
      {
        mag_[numFreq_*k + j] += (newValues_[gridSize_*k+i]*sin(j*2.0*M_PI*i/((double) gridSize_)));
        phase_[numFreq_*k + j] += (newValues_[gridSize_*k+i]*cos(j*2.0*M_PI*i/((double) gridSize_)));
      }
    }
  
    mag_[numFreq_*k] = phase_[numFreq_*k]/gridSize_;
    phase_[numFreq_*k] = 0;
    thd_[k] = 0; 

    double convRadDeg = 180.0/M_PI;

    for(int i = 1; i < numFreq_ ; i++)
    { 
      double tmp = mag_[numFreq_*k+i]*2.0 /gridSize_;
      phase_[numFreq_*k+i] *= 2.0/gridSize_;
      mag_[numFreq_*k+i] = sqrt(tmp*tmp+phase_[numFreq_*k+i]*phase_[numFreq_*k+i]);
      phase_[numFreq_*k+i] = atan2(phase_[numFreq_*k+i],tmp)*convRadDeg;
      nmag_[numFreq_*k+i] = mag_[numFreq_*k+i]/mag_[numFreq_*k+1];
      nphase_[numFreq_*k+i] = phase_[numFreq_*k+i]-phase_[numFreq_*k+1];
      if(i>1) thd_[k] += nmag_[numFreq_*k+i]*nmag_[numFreq_*k+i];
    }
    thd_[k] = 100*sqrt(thd_[k]);
  }
}

//-----------------------------------------------------------------------------
// Function      : FourierMgr::printResult_( std::ostream& os )
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, SNL, Electrical Models & Simulation
// Creation Date : 6/21/13
//-----------------------------------------------------------------------------
std::ostream& FourierMgr::printResult_( std::ostream& os )
{ 
  // Compute measure.
  if (calculated_)
  { 
    for (unsigned int i=0; i<freqVector_.size(); ++i)
    {
      for (int j=outputVarsPtr_[i]; j<outputVarsPtr_[i+1]; j++)
      {
        int colWidth1=12, colWidth2 = 16, precision = 6;
        os << "Fourier analysis for " << names_[j] << ":" << std::endl;
        os << "  No. Harmonics: " << numFreq_ << ", THD: " 
           << std::scientific << std::setprecision(precision)
           << thd_[j] << ", Gridsize: " << gridSize_
           << ", Interpolation Type: Cubic Spline" << std::endl;
  
        os << std::setw(colWidth1) << "Harmonic" << std::setw(colWidth2) << "Frequency"
           << std::setw(colWidth2) << "Magnitude" << std::setw(colWidth2) << "Phase"
           << std::setw(colWidth2) << "Norm. Mag" << std::setw(colWidth2) << "Norm. Phase" << std::endl;
        for (int k = 0; k < numFreq_; k++)
        {
           os << std::setw(colWidth1) << k << std::setw(colWidth2) << freq_[numFreq_*i+k]
              << std::setw(colWidth2) << mag_[numFreq_*j+k]
              << std::setw(colWidth2) << phase_[numFreq_*j+k]
              << std::setw(colWidth2) << nmag_[numFreq_*j+k]
              << std::setw(colWidth2) << nphase_[numFreq_*j+k] << std::endl;
        }
        os << std::endl;
      }  
    }
  }
  return os;
}


namespace {

//-----------------------------------------------------------------------------
// Function      : extractFOURIERData
// Purpose       : Convert a .four line to an options block
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist
// Creation Date : 06/03/2013
//-----------------------------------------------------------------------------
bool
extractFOURIERData(
  IO::PkgOptionsMgr &           options_manager,
  IO::CircuitBlock &            circuit_block,
  const std::string &           netlist_filename,
  const IO::TokenVector &       parsed_line)
{
  Util::OptionBlock option_block("FOUR", Util::OptionBlock::ALLOW_EXPRESSIONS, netlist_filename, parsed_line[0].lineNumber_);

  Util::Param parameter;
  ExtendedString nextWord("");

  if(parsed_line.size() < 3)
  {
    Report::UserError0().at(netlist_filename, parsed_line[0].lineNumber_)
      << "Error: .FOUR line needs at least 3 arguments '.FOUR freq ov1 <ov2 ...>'";
    return false;
  }

  parameter.setTag("FREQ");
  parameter.setVal(parsed_line[1].string_ );
  option_block.addParam( parameter );

  int position = 2;
  int endPosition = parsed_line.size();
  while (position < endPosition)
  {

    nextWord = parsed_line[position].string_;
    nextWord.toUpper();
    // nextWord[0] is used for I because branch currents for transistors can have two
    // characters.  An example is IS for the M-Device.
    if( (nextWord[0] == 'I') || (nextWord == "V") || (nextWord == "P") || (nextWord == "W") || (nextWord == "N") )
    {
      // need to do a bit of look ahead here to see if this is a V(a) or V(a,b)
      int numNodes = 1;
      if( ((position+3) < endPosition) && (parsed_line[position+3].string_ == ",") )
      {
        numNodes = 2;
      }
      parameter.set(nextWord, numNodes);
      option_block.addParam( parameter );

      if( (position+2) < endPosition )
      {
        position+=2;
        nextWord = parsed_line[position].string_;
        nextWord.toUpper();
        parameter.set( nextWord, 0.0 );
        option_block.addParam( parameter );

        if( ((position+2) < endPosition) && (parsed_line[position+1].string_ == ",") )
        {
          // may have a voltage difference request
          position+=2;
          nextWord = parsed_line[position].string_;
          nextWord.toUpper();
          parameter.set( nextWord, 0.0 );
          option_block.addParam( parameter );
        }
      }
      else
      {
        Report::UserError0().at(netlist_filename, parsed_line[position].lineNumber_) << "Could not parse .FOUR variable";
      }
      position+=2;
    }
    else if ( (nextWord[0]=='{') && (nextWord[nextWord.size()-1]=='}') )
    {
      parameter.set(nextWord, std::string(nextWord));
      option_block.addParam(parameter);
      position++;
    }
    else
    {
      Report::UserError().at(netlist_filename, parsed_line[position].lineNumber_) << "Could not parse .FOUR variable " << nextWord;
      position+=2;
    }
  }

  circuit_block.addOptions(option_block);

  return true;
}

} // namespace <unnamed>

//-----------------------------------------------------------------------------
// Function      : OutputMgr::registerPkgOptionsMgr
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, 1437
// Creation Date : 10/21/08
//-----------------------------------------------------------------------------
bool registerPkgOptionsMgr(FourierMgr &fourier_manager, PkgOptionsMgr &options_manager)
{
  options_manager.addCommandParser(".FOUR", extractFOURIERData);

  options_manager.addCommandProcessor("FOUR", IO::createRegistrationOptions(fourier_manager, &FourierMgr::addFourierAnalysis));
  return true;
}

} // namespace IO
} // namespace Xyce
