//-------------------------------------------------------------------------
//   Copyright 2002-2018 National Technology & Engineering Solutions of
//   Sandia, LLC (NTESS).  Under the terms of Contract DE-NA0003525 with
//   NTESS, the U.S. Government retains certain rights in this software.
//
//   This file is part of the Xyce(TM) Parallel Electrical Simulator.
//
//   Xyce(TM) is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   Xyce(TM) is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with Xyce(TM).
//   If not, see <http://www.gnu.org/licenses/>.
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
//
// Purpose        : Build various operators, like V(), I(), N() and P()
//
// Special Notes  :
//
// Creator        : Dave Baur
//
// Creation Date  : 08/04/14
//
//
//
//
//-------------------------------------------------------------------------

#include <Xyce_config.h>

#include <string>
#include <vector>

#include <N_IO_fwd.h>
#include <N_UTL_fwd.h>

#include <N_ERH_ErrorMgr.h>
#include <N_IO_MeasureManager.h>
#include <N_IO_Op.h>
#include <N_UTL_Marshal.h>
#include <N_UTL_OpBuilder.h>
#include <N_UTL_Param.h>
#include <N_UTL_DeviceNameConverters.h>
#include <N_UTL_FeatureTest.h>

namespace Xyce {
namespace IO {

namespace {

//-------------------------------------------------------------------------- 
// Function      : parameterNameAndArgs 
// Purpose       : Return a parameter's names and arguments
// Special Notes :  
// Creator       : Dave Baur
// Creation Date : 08/04/14 
//--------------------------------------------------------------------------
void parameterNameAndArgs(
  std::string &                         name,
  std::vector<std::string> &            args,
  Util::ParamList::const_iterator &     it)
{
  const std::string &param_tag = (*it).tag();

  if ((*it).getType() == Util::INT && (param_tag[0] == 'V' || param_tag[0] == 'I' || param_tag[0] == 'N' || 
                                       param_tag[0] == 'P' || param_tag[0] == 'W' || param_tag[0] == 'D'))
  {
    std::ostringstream oss;
    oss << param_tag << "(";
    int arg_count = (*it).getImmutableValue<int>();
    for (int i = 0; i < arg_count; ++i)
    {
      ++it;
      if (i != 0)
        oss << ",";
      oss << (*it).tag();
      args.push_back((*it).tag());
    }
    oss << ")";
    name = oss.str();
  }
}

//-------------------------------------------------------------------------- 
// Function      : findNode 
// Purpose       : Determine whether a requested node name is "valid".
//                 That includes node names in the solution vector, the
//                 aliasNodeMap and also Ground (0 or GND).
// Special Notes :  
// Creator       : Dave Baur
// Creation Date : 08/04/14 
//--------------------------------------------------------------------------
NodeNameMap::const_iterator findNode(
  const std::string &           name,
  const NodeNameMap &       node_map,
  const AliasNodeMap &          alias_map)
{
  NodeNameMap::const_iterator node_it = node_map.find(name);
  if (node_it == node_map.end()) 
  {
    AliasNodeMap::const_iterator alias_node_it = alias_map.find(name);
    if (alias_node_it != alias_map.end())
      node_it = node_map.find((*alias_node_it).second);
  }

  return node_it;
}

} // namespace <unnamed>

//-------------------------------------------------------------------------- 
// Structure     : Util::Op::Builder::CircuitTemperatureOpBuilder 
// Purpose       : This creates an OutputMgrTemperatureOp
// Special Notes :  
// Creator       : Dave Baur
// Creation Date : 08/04/14 
//--------------------------------------------------------------------------
struct CircuitTemperatureOpBuilder : public Util::Op::Builder
{
  CircuitTemperatureOpBuilder(const OutputMgr & output_manager)
    : outputManager_(output_manager)
  {}

  virtual ~CircuitTemperatureOpBuilder()
  {}

  virtual void registerCreateFunctions(Util::Op::BuilderManager &builder_manager) const
  {
    builder_manager.addCreateFunction<OutputMgrTemperatureOp>();
  }

  virtual Util::Op::Operator *makeOp(Util::ParamList::const_iterator &it) const 
  {
    Util::Op::Operator *new_op = 0;
    const std::string &param_tag = (*it).tag();

    if (param_tag == "TEMP") 
    {
      new_op  = new OutputMgrTemperatureOp(param_tag, outputManager_);
    }

    return new_op;
  }

private:
  const OutputMgr &     outputManager_;
};

//-------------------------------------------------------------------------- 
// Structure     : Util::Op::Builder::CircuitTimeOpBuilder 
// Purpose       : This creates an OutputMgrTimeOp
// Special Notes :  
// Creator       : Dave Baur
// Creation Date : 08/04/14 
//--------------------------------------------------------------------------
struct CircuitTimeOpBuilder : public Util::Op::Builder
{
  CircuitTimeOpBuilder(const OutputMgr & output_manager)
    : outputManager_(output_manager)
  {}

  virtual ~CircuitTimeOpBuilder()
  {}

  virtual void registerCreateFunctions(Util::Op::BuilderManager &builder_manager) const
  {
    builder_manager.addCreateFunction<OutputMgrTimeOp>();
  }

  virtual Util::Op::Operator *makeOp(Util::ParamList::const_iterator &it) const 
  {
    Util::Op::Operator *new_op = 0;
    const std::string &param_tag = (*it).tag();

    if (param_tag == "TIME") 
    {
      new_op  = new OutputMgrTimeOp(param_tag, outputManager_, 1.0);
    }

    return new_op;
  }

private:
  const OutputMgr &     outputManager_;
};

//-------------------------------------------------------------------------- 
// Structure     : Util::Op::Builder::CircuitOutputNoiseOpBuilder 
// Purpose       : This creates an OutputMgrOutputNoiseOp
// Special Notes : 
// Creator       : Dave Baur
// Creation Date : 08/04/14 
//--------------------------------------------------------------------------
struct CircuitOutputNoiseOpBuilder : public Util::Op::Builder
{
  CircuitOutputNoiseOpBuilder(const OutputMgr & output_manager)
    : outputManager_(output_manager)
  {}

  virtual ~CircuitOutputNoiseOpBuilder()
  {}

  virtual void registerCreateFunctions(Util::Op::BuilderManager &builder_manager) const
  {
    builder_manager.addCreateFunction<OutputMgrOutputNoiseOp>();
  }

  virtual Util::Op::Operator *makeOp(Util::ParamList::const_iterator &it) const 
  {
    Util::Op::Operator *new_op = 0;
    const std::string &param_tag = (*it).tag();

    if (param_tag == "ONOISE") 
    {
      new_op  = new OutputMgrOutputNoiseOp(param_tag, outputManager_);
    }

    return new_op;
  }

private:
  const OutputMgr &     outputManager_;
};

//-------------------------------------------------------------------------- 
// Structure     : Util::Op::Builder::CircuitInputNoiseOpBuilder 
// Purpose       : This creates an OutputMgrInputNoiseOp 
// Special Notes : 
// Creator       : Dave Baur
// Creation Date : 08/04/14 
//--------------------------------------------------------------------------
struct CircuitInputNoiseOpBuilder : public Util::Op::Builder
{
  CircuitInputNoiseOpBuilder(const OutputMgr & output_manager)
    : outputManager_(output_manager)
  {}

  virtual ~CircuitInputNoiseOpBuilder()
  {}

  virtual void registerCreateFunctions(Util::Op::BuilderManager &builder_manager) const
  {
    builder_manager.addCreateFunction<OutputMgrInputNoiseOp>();
  }

  virtual Util::Op::Operator *makeOp(Util::ParamList::const_iterator &it) const 
  {
    Util::Op::Operator *new_op = 0;
    const std::string &param_tag = (*it).tag();

    if (param_tag == "INOISE") 
    {
      new_op  = new OutputMgrInputNoiseOp(param_tag, outputManager_);
    }

    return new_op;
  }

private:
  const OutputMgr &     outputManager_;
};

//-------------------------------------------------------------------------- 
// Structure     : Util::Op::Builder::CircuitNoiseContOpBuilder 
// Purpose       : This creates an either an OutputMgrOutputNoiseContOp
//               : or an OutputMgrInputNoiseContOp
// Special Notes : 
// Creator       : Pete Sholander
// Creation Date : 11/20/17 
//--------------------------------------------------------------------------
struct CircuitNoiseContOpBuilder : public Util::Op::Builder
{
  CircuitNoiseContOpBuilder(const OutputMgr & output_manager)
    : outputManager_(output_manager)
  {}

  virtual ~CircuitNoiseContOpBuilder()
  {}

  virtual void registerCreateFunctions(Util::Op::BuilderManager &builder_manager) const
  {
    builder_manager.addCreateFunction<OutputMgrOutputNoiseContOp>();
    builder_manager.addCreateFunction<OutputMgrInputNoiseContOp>();
  }

  virtual Util::Op::Operator *makeOp(Util::ParamList::const_iterator &it) const 
  {
    Util::Op::Operator *new_op = 0;
    const std::string &param_tag = (*it).tag();

    std::vector<std::string> args;
    std::string name;
    parameterNameAndArgs(name, args, it);

    if (param_tag[0] == 'D' && args.size() > 0) 
    {
      // find the index of the noise source, and then of the noise type if the source
      // has multiple noise types.
      int srcIndex = -1;
      int typeIndex = -1;
      NodeNameMap::const_iterator ns_it =  outputManager_.getNoiseDataVecNodeMap().find(args[0]+"_NS");
      if (ns_it != outputManager_.getNoiseDataVecNodeMap().end())
           srcIndex = (*ns_it).second;

      // Now handle the noise type, if one was specified
      if (args.size() == 2)
      {
	NodeNameMap::const_iterator nt_it = outputManager_.getNoiseDataVecNodeMap().find("noise_" + args[0] + "_" + args[1]);
        NodeNameMap::const_iterator nt_it_adms = outputManager_.getNoiseDataVecNodeMap().find(args[0] + "_onoise_" + args[1]);
        if (nt_it != outputManager_.getNoiseDataVecNodeMap().end())
	{
          typeIndex = (*nt_it).second;
        }
        else if (nt_it_adms != outputManager_.getNoiseDataVecNodeMap().end())
	{
	   // temporary code, since the ADMS devices still use onoise, and the position of the device
           // name is different than in the open models. 
           typeIndex = (*nt_it_adms).second;
        }
      }  
      
      // Only make the op if the relevant indices have been found.
      // If not then this op will be flagged as in error, since new_op is still a null pointer 
      if ( (srcIndex != -1) && (( args.size() == 1 ) || ((args.size() == 2) && (typeIndex != -1))) )
      {
        if (param_tag == "DNI")
        {
          new_op  = new OutputMgrInputNoiseContOp(name, srcIndex, typeIndex, outputManager_);
        }
        else if (param_tag == "DNO")
        {
          new_op  = new OutputMgrOutputNoiseContOp(name, srcIndex, typeIndex, outputManager_);
        }
      }
    }

    if (new_op)
          new_op->addArg(args[0]);
    
    return new_op;
  }

private:
  const OutputMgr &     outputManager_;
};

//-------------------------------------------------------------------------- 
// Structure     : Util::Op::Builder::CircuitFrequencyOpBuilder 
// Purpose       : This creates an OutputMgrFrequencyOp 
// Special Notes : 
// Creator       : Dave Baur
// Creation Date : 08/04/14 
//--------------------------------------------------------------------------
struct CircuitFrequencyOpBuilder : public Util::Op::Builder
{
  CircuitFrequencyOpBuilder(const OutputMgr & output_manager)
    : outputManager_(output_manager)
  {}

  virtual ~CircuitFrequencyOpBuilder()
  {}

  virtual void registerCreateFunctions(Util::Op::BuilderManager &builder_manager) const
  {
    builder_manager.addCreateFunction<OutputMgrFrequencyOp>();
  }

  virtual Util::Op::Operator *makeOp(Util::ParamList::const_iterator &it) const {
    Util::Op::Operator *new_op = 0;
    const std::string &param_tag = (*it).tag();

    if (param_tag == "FREQUENCY") {
      new_op  = new OutputMgrFrequencyOp(param_tag, outputManager_);
    }

    return new_op;
  }

private:
  const OutputMgr &     outputManager_;
};

//-------------------------------------------------------------------------- 
// Structure     : Util::Op::Builder::StepSweepOpBuilder 
// Purpose       : This creates an OutputMgrStepSweepOp 
// Special Notes : 
// Creator       : Dave Baur
// Creation Date : 08/04/14 
//--------------------------------------------------------------------------
struct StepSweepOpBuilder : public Util::Op::Builder
{
  StepSweepOpBuilder(const OutputMgr & output_manager)
    : outputManager_(output_manager)
  {}

  virtual ~StepSweepOpBuilder()
  {}

  virtual void registerCreateFunctions(Util::Op::BuilderManager &builder_manager) const
  {
    builder_manager.addCreateFunction<OutputMgrStepSweepOp>();
  }

  virtual Util::Op::Operator *makeOp(Util::ParamList::const_iterator &it) const 
  {
    Util::Op::Operator *new_op = 0;
    const std::string &param_tag = (*it).tag();

    for (size_t i = 0; i < outputManager_.getStepSweepVector().size(); ++i)
    {
      if (param_tag == outputManager_.getStepSweepVector()[i].name)
      {
        new_op = new OutputMgrStepSweepOp(param_tag, outputManager_, i);
        break;
      }
    }

    return new_op;
  }

private:
  const OutputMgr &     outputManager_;
};

//-------------------------------------------------------------------------- 
// Structure     : Util::Op::Builder::DCSweepOpBuilder 
// Purpose       : This creates an OutputMgrDCSweepOp 
// Special Notes : 
// Creator       : Dave Baur
// Creation Date : 08/04/14 
//--------------------------------------------------------------------------
struct DCSweepOpBuilder : public Util::Op::Builder
{
  DCSweepOpBuilder(const OutputMgr & output_manager)
    : outputManager_(output_manager)
  {}

  virtual ~DCSweepOpBuilder()
  {}

  virtual void registerCreateFunctions(Util::Op::BuilderManager &builder_manager) const
  {
    builder_manager.addCreateFunction<OutputMgrDCSweepOp>();
  }

  virtual Util::Op::Operator *makeOp(Util::ParamList::const_iterator &it) const 
  {
    Util::Op::Operator *new_op = 0;
    const std::string &param_tag = (*it).tag();

    for (size_t i = 0; i < outputManager_.getDCSweepVector().size(); ++i)
    {
      if (param_tag == outputManager_.getDCSweepVector()[i].name)
      {
        new_op = new OutputMgrDCSweepOp(param_tag, outputManager_, i);
        break;
      }
    }

    return new_op;
  }

private:
  const OutputMgr &     outputManager_;
};

//-------------------------------------------------------------------------- 
// Structure     : Util::Op::Builder::DCSweepCurrentValueOpBuilder 
// Purpose       : This creates an OutputMgrDCSweepCurrentValueOp 
// Special Notes : 
// Creator       : Dave Baur
// Creation Date : 08/04/14 
//--------------------------------------------------------------------------
struct DCSweepCurrentValueOpBuilder : public Util::Op::Builder
{
  DCSweepCurrentValueOpBuilder(const OutputMgr & output_manager)
    : outputManager_(output_manager)
  {}

  virtual ~DCSweepCurrentValueOpBuilder()
  {}

  virtual void registerCreateFunctions(Util::Op::BuilderManager &builder_manager) const
  {
    builder_manager.addCreateFunction<OutputMgrDCSweepCurrentValueOp>();
  }

  virtual Util::Op::Operator *makeOp(Util::ParamList::const_iterator &it) const 
  {
    Util::Op::Operator *new_op = 0;
    const std::string &param_tag = (*it).tag();

    if (param_tag == "sweep") {
      new_op  = new OutputMgrDCSweepCurrentValueOp(param_tag, outputManager_);
    }

    return new_op;
  }

private:
  const OutputMgr &     outputManager_;
};

//-------------------------------------------------------------------------- 
// Structure     : Util::Op::Builder::InternalVariableOpBuilder 
// Purpose       : Builds various "solution operators".  See the list
//                 below in registerCreateFunctions().  It includes N(), 
//                 StateOp() and StoreOp().  The code below also lists
//                 NR(), NI(), NM(), NP() and NDB(). However, those operators
//                 won't pass netlist parsing and aren't fully implemented
//                 yet. See SON Bug 880 for more details. 
// Special Notes : 
// Creator       : Dave Baur
// Creation Date : 08/04/14 
//--------------------------------------------------------------------------
struct InternalVariableOpBuilder : public Util::Op::Builder
{
  InternalVariableOpBuilder(const OutputMgr & output_manager)
    : outputManager_(output_manager)
  {}

  virtual ~InternalVariableOpBuilder()
  {}

  virtual void registerCreateFunctions(Util::Op::BuilderManager &builder_manager) const
  {
    builder_manager.addCreateFunction<SolutionOp>();
    builder_manager.addCreateFunction<SolutionImaginaryOp>();
    builder_manager.addCreateFunction<SolutionMagnitudeOp>();
    builder_manager.addCreateFunction<SolutionPhaseOp>();
    builder_manager.addCreateFunction<SolutionDecibelsOp>();
    builder_manager.addCreateFunction<StateOp>();
    builder_manager.addCreateFunction<StoreOp>();
  }

  virtual Util::Op::Operator *makeOp(Util::ParamList::const_iterator &it) const 
  {
    Util::Op::Operator *new_op = 0;
    const std::string &param_tag = (*it).tag();

    std::vector<std::string> args;
    std::string name;
    parameterNameAndArgs(name, args, it);

    if (param_tag[0] == 'N' && args.size() == 1)
    {
      NodeNameMap::const_iterator it = findNode(args[0], outputManager_.getSolutionNodeMap(), outputManager_.getAliasNodeMap());
      if (it != outputManager_.getSolutionNodeMap().end())
      {
        int index = (*it).second;
        if (param_tag == "N" )
        {
          new_op = new SolutionOp(name, index);
        }
        else if (param_tag == "NR" )
        {
          new_op = new SolutionRealOp(name, index);
        }
        else if (param_tag == "NI" )
        {
          new_op = new SolutionImaginaryOp(name, index);
        }
        else if (param_tag == "NM" )
        {
          new_op = new SolutionMagnitudeOp(name, index);
        }
        else if (param_tag == "NP" )
        {
          new_op = new SolutionPhaseOp(name, index);
        }
        else if (param_tag == "NDB" )
        {
          new_op = new SolutionDecibelsOp(name, index);
        }
      }
      else
      {
        it = findNode(args[0], outputManager_.getStateNodeMap(), outputManager_.getAliasNodeMap());
        if (it != outputManager_.getStateNodeMap().end())
        {
          new_op = new StateOp(name, (*it).second);
        }
        else
        {
          it = outputManager_.getStoreNodeMap().find(args[0]);
          if (it != outputManager_.getStoreNodeMap().end())
          {
            new_op = new StoreOp(name, (*it).second);
          }
          else
          {
            new_op = new Util::Op::UndefinedOp(param_tag);
          }
        }
      }

      if (new_op)
        new_op->addArg(args[0]);
    }

    return new_op;
  }

private:
  const OutputMgr &     outputManager_;
};

//-------------------------------------------------------------------------- 
// Structure     : Util::Op::Builder::VoltageVariableOpBuilder 
// Purpose       : Builds various "solution operators" for voltage variables.
//                 See the list below in registerCreateFunctions().  This
//                 includes the operators for V(), VR(), VI(), VM(), VP()
//                 and VDB(). It includes, for example, V(a) and also the
//                 corresponding "voltage difference" form of V(a,b).
// Special Notes : 
// Creator       : Dave Baur
// Creation Date : 08/04/14 
//--------------------------------------------------------------------------
struct VoltageVariableOpBuilder : public Util::Op::Builder
{
  VoltageVariableOpBuilder(const OutputMgr & output_manager)
    : outputManager_(output_manager)
  {}

  virtual ~VoltageVariableOpBuilder()
  {}

  virtual void registerCreateFunctions(Util::Op::BuilderManager &builder_manager) const
  {
    builder_manager.addCreateFunction<SolutionOp>();
    builder_manager.addCreateFunction<SolutionRealOp>();
    builder_manager.addCreateFunction<SolutionImaginaryOp>();
    builder_manager.addCreateFunction<SolutionMagnitudeOp>();
    builder_manager.addCreateFunction<SolutionPhaseOp>();
    builder_manager.addCreateFunction<SolutionDecibelsOp>();
    builder_manager.addCreateFunction<VoltageDifferenceOp>();
    builder_manager.addCreateFunction<VoltageDifferenceRealOp>();
    builder_manager.addCreateFunction<VoltageDifferenceImaginaryOp>();
    builder_manager.addCreateFunction<VoltageDifferenceMagnitudeOp>();
    builder_manager.addCreateFunction<VoltageDifferencePhaseOp>();
    builder_manager.addCreateFunction<VoltageDifferenceDecibelsOp>();
  }

  virtual Util::Op::Operator *makeOp(Util::ParamList::const_iterator &it) const 
  {
    Util::Op::Operator *new_op = 0;
    const std::string &param_tag = (*it).tag();

    std::vector<std::string> args;
    std::string name;
    parameterNameAndArgs(name, args, it);

    if (param_tag[0] == 'V' && args.size() > 0)
    {

      // Solution variable
      if (args.size() == 1)
      {
        int index = -1;
        NodeNameMap::const_iterator it = findNode(args[0], outputManager_.getSolutionNodeMap(), outputManager_.getAliasNodeMap());
        if (it != outputManager_.getSolutionNodeMap().end())
          index = (*it).second;

        if (param_tag == "V" )
        {
          new_op = new SolutionOp(name, index);
        }
        else if (param_tag == "VR" )
        {
          new_op = new SolutionRealOp(name, index);
        }
        else if (param_tag == "VI" )
        {
          new_op = new SolutionImaginaryOp(name, index);
        }
        else if (param_tag == "VM" )
        {
          new_op = new SolutionMagnitudeOp(name, index);
        }
        else if (param_tag == "VP" )
        {
          new_op = new SolutionPhaseOp(name, index);
        }
        else if (param_tag == "VDB" )
        {
          new_op = new SolutionDecibelsOp(name, index);
        }

        if (new_op)
          new_op->addArg(args[0]);
      }

      // Voltage Difference
      if (args.size() == 2)
      {
        int index1 = -1;
        int index2 = -1;

        NodeNameMap::const_iterator it = findNode(args[0], outputManager_.getSolutionNodeMap(), outputManager_.getAliasNodeMap());
        if (it != outputManager_.getSolutionNodeMap().end())
          index1 = (*it).second;

        it = findNode(args[1], outputManager_.getSolutionNodeMap(), outputManager_.getAliasNodeMap());
        if (it != outputManager_.getSolutionNodeMap().end())
          index2 = (*it).second;

        if (param_tag == "V" )
        {
          new_op = new VoltageDifferenceOp(name, index1, index2);
        }
        else if (param_tag == "VR" )
        {
          new_op = new VoltageDifferenceRealOp(name, index1, index2);
        }
        else if (param_tag == "VI" )
        {
          new_op = new VoltageDifferenceImaginaryOp(name, index1, index2);
        }
        else if (param_tag == "VM" )
        {
          new_op = new VoltageDifferenceMagnitudeOp(name, index1, index2);
        }
        else if (param_tag == "VP" )
        {
          new_op = new VoltageDifferencePhaseOp(name, index1, index2);
        }
        else if (param_tag == "VDB" )
        {
          new_op = new VoltageDifferenceDecibelsOp(name, index1, index2);
        }

        if (new_op)
          new_op->addArgs(args.begin(), args.end());
      }
    }

    return new_op;
  }

private:
  const OutputMgr &     outputManager_;
};

//-------------------------------------------------------------------------- 
// Allowed set of current operators, besides I()
// Creator       : Dave Baur
// Creation Date : 08/04/14 
//--------------------------------------------------------------------------
namespace {
static const char * const func_names[] = {"II", "IR", "IP", "IM", "IDB"};
}

//-------------------------------------------------------------------------- 
// Structure     : Util::Op::Builder::CurrentVariableOpBuilder 
// Purpose       : Builds for various "solution operators" and "store
//                 operators" for lead current variables.  See the list
//                 below in registerCreateFunctions().  This includes the 
//                 operators for I(), IR(), II(), IM(), IP() and IDB(). 
//                 The BranchData forms are used to get values for lead 
//                 currents that are not part of the solution vector.
//                 Note also that special handling is needed for YPDE, C 
//                 and L devices.
// Special Notes :
// Creator       : Dave Baur
// Creation Date : 08/04/14 
//--------------------------------------------------------------------------
struct CurrentVariableOpBuilder : public Util::Op::Builder
{

  CurrentVariableOpBuilder(const OutputMgr & output_manager)
    : outputManager_(output_manager)
  {}

  virtual ~CurrentVariableOpBuilder()
  {}

  virtual void registerCreateFunctions(Util::Op::BuilderManager &builder_manager) const
  {
    builder_manager.addCreateFunction<SolutionOp>();
    builder_manager.addCreateFunction<SolutionRealOp>();
    builder_manager.addCreateFunction<SolutionImaginaryOp>();
    builder_manager.addCreateFunction<SolutionMagnitudeOp>();
    builder_manager.addCreateFunction<SolutionPhaseOp>();
    builder_manager.addCreateFunction<SolutionDecibelsOp>();
    builder_manager.addCreateFunction<StoreOp>();
    builder_manager.addCreateFunction<StoreRealOp>();
    builder_manager.addCreateFunction<StoreImaginaryOp>();
    builder_manager.addCreateFunction<StoreMagnitudeOp>();
    builder_manager.addCreateFunction<StorePhaseOp>();
    builder_manager.addCreateFunction<StoreDecibelsOp>();
    builder_manager.addCreateFunction<BranchDataCurrentOp>();
    builder_manager.addCreateFunction<BranchDataCurrentRealOp >();
    builder_manager.addCreateFunction<BranchDataCurrentImaginaryOp>();
    builder_manager.addCreateFunction<BranchDataCurrentMagnitudeOp>();
    builder_manager.addCreateFunction<BranchDataCurrentPhaseOp>();
    builder_manager.addCreateFunction<BranchDataCurrentDecibelsOp>();
  }

  virtual Util::Op::Operator *makeOp(Util::ParamList::const_iterator &it) const 
  {
    Util::Op::Operator *new_op = 0;
    const std::string &param_tag = (*it).tag();

    std::vector<std::string> args;
    std::string name;
    parameterNameAndArgs(name, args, it);

    if (param_tag[0] == 'I' && !args.empty())
    {
      // Node name could be circuit_context:DeviceTypeDeviceName while internally it should be DeviceType:circuit_context:DeviceName.
      std::string modifiedName = Util::xyceDeviceNameToSpiceName(args[0]);

      // could be a device lead current "DEV_I" or a branch current.
      // so we don't have to duplicate solution vars(branch currents) in the
      // store vector, look for each type.
      bool param_func = std::find(func_names, func_names + sizeof(func_names)/sizeof(func_names[0]), param_tag) != func_names + sizeof(func_names)/sizeof(func_names[0]);
      std::string store_name = modifiedName + ":DEV_" + (param_func ? "I" : param_tag);  // if it is in the state/store vec.
      std::string leadCurrent_name = modifiedName + ":BRANCH_D";
      if (!param_func && (param_tag.length() > 1) )
      {
        leadCurrent_name += param_tag[1];
      } 

      // this if block allows for spaces in YPDE names as in I1(YPDE NAME)
      // we try to find devices based on store_name in the following blocks of code,
      // so do this modification now.
      std::string::size_type space = store_name.find_first_of(" ");
      if (space != std::string::npos)
      {
        if (space == 4 && store_name.substr(0, 4) == "YPDE")
        {
          store_name.replace(4, 1, ":");
        }
      }
      
      NodeNameMap::const_iterator it;
      // Search solution vector first, specifically excluding the C Device.  For the C device, 
      // we will search in the lead current vector.  If the IC parameter is specified for a
      // capacitor, then a voltage source is used to enforce that IC value at the DCOP.  However,
      // the contribution from that V-source has been included in the lead-current vector.  
      if ( !new_op && !startswith_nocase(modifiedName, "C") ) 
      {
        std::string solution_name = modifiedName + "_BRANCH";         // if it is in the solution vec.
        const NodeNameMap &x = outputManager_.getSolutionNodeMap();
        it = x.find(solution_name);
        if (it != outputManager_.getSolutionNodeMap().end())
        {
          int index = (*it).second;
          if (param_tag == "I" )
          {
            new_op = new SolutionOp(name, index);
          }
          else if (param_tag == "IR" )
          {
            new_op = new SolutionRealOp(name, index);
          }
          else if (param_tag == "II" )
          {
            new_op = new SolutionImaginaryOp(name, index);
          }
          else if (param_tag == "IM" )
          {
            new_op = new SolutionMagnitudeOp(name, index);
          }
          else if (param_tag == "IP" )
          {
            new_op = new SolutionPhaseOp(name, index);
          }
          else if (param_tag == "IDB" )
          {
            new_op = new SolutionDecibelsOp(name, index);
          }
        }
      }

      // Search solution
      if (!new_op && startswith_nocase(modifiedName, "L")) // Mutual inductor special name
      {
        const NodeNameMap &x = outputManager_.getSolutionNodeMap();
        it = x.find(modifiedName);
        if (it != outputManager_.getSolutionNodeMap().end())
        {
          int index = (*it).second;
          if (param_tag == "I" )
          {
            new_op = new SolutionOp(name, index);
          }
          else if (param_tag == "IR" )
          {
            new_op = new SolutionRealOp(name, index);
          }
          else if (param_tag == "II" )
          {
            new_op = new SolutionImaginaryOp(name, index);
          }
          else if (param_tag == "IM" )
          {
            new_op = new SolutionMagnitudeOp(name, index);
          }
          else if (param_tag == "IP" )
          {
            new_op = new SolutionPhaseOp(name, index);
          }
          else if (param_tag == "IDB" )
          {
            new_op = new SolutionDecibelsOp(name, index);
          }
        }
      }

      if( !new_op ) 
      {
        it = outputManager_.getBranchVarsNodeMap().find(leadCurrent_name );
        // Search lead current vector 
        if (it != outputManager_.getBranchVarsNodeMap().end())
        {
          int index = (*it).second;
          //new_op = new BranchDataCurrentOp(name, index);
            // need to support the following for other lead current usage 
          if (param_tag == "IR" )
          {
            new_op = new BranchDataCurrentRealOp(name, index);
          }
          else if (param_tag == "II" )
          {
            new_op = new BranchDataCurrentImaginaryOp(name, index);
          }
          else if (param_tag == "IM" )
          {
            new_op = new BranchDataCurrentMagnitudeOp(name, index);
          }
          else if (param_tag == "IP" )
          {
            new_op = new BranchDataCurrentPhaseOp(name, index);
          }
          else if (param_tag == "IDB" )
          {
            new_op = new BranchDataCurrentDecibelsOp(name, index);
          }
          else // IC, IE, IB
          {
            new_op = new BranchDataCurrentOp(name, index);
          }
        }
      } 
 
      if( !new_op )
      {
        // Search store
        NodeNameMap::const_iterator it = outputManager_.getStoreNodeMap().find(store_name);
        if (it != outputManager_.getStoreNodeMap().end())
        {
          int index = (*it).second;
          if (param_tag == "IR" )
          {
            new_op = new StoreRealOp(name, index);
          }
          else if (param_tag == "II" )
          {
            new_op = new StoreImaginaryOp(name, index);
          }
          else if (param_tag == "IM" )
          {
            new_op = new StoreMagnitudeOp(name, index);
          }
          else if (param_tag == "IP" )
          {
            new_op = new StorePhaseOp(name, index);
          }
          else if (param_tag == "IDB" )
          {
            new_op = new StoreDecibelsOp(name, index);
          }
          else // IC, IE, IB
          {
            new_op = new StoreOp(name, index);
          }
        }
      }
 
      if (new_op)
      {
        new_op->addArg(args[0]);
      }
    }

    return new_op;
  }

private:
  const OutputMgr &     outputManager_;
};

//-------------------------------------------------------------------------- 
// Structure     : Util::Op::Builder::PowerVariableOpBuilder 
// Purpose       : Builds various operators" for power calculations.  See 
//                 the list below in registerCreateFunctions(). This includes  
//                 includes the operators for P() and W().  Note that
//                 different operators are needed for each of the various 
//                 multi-terminal devices like J, Q, M, T and Z.
// Special Notes :
// Creator       : Dave Baur
// Creation Date : 08/04/14 
//--------------------------------------------------------------------------
struct PowerVariableOpBuilder : public Util::Op::Builder
{

  PowerVariableOpBuilder(const OutputMgr & output_manager)
    : outputManager_(output_manager)
  {}

  virtual ~PowerVariableOpBuilder()
  {}

  virtual void registerCreateFunctions(Util::Op::BuilderManager &builder_manager) const
  {
    builder_manager.addCreateFunction<BranchDataCurrentOp>();
    builder_manager.addCreateFunction<BranchDataCurrentRealOp >();
    builder_manager.addCreateFunction<BranchDataCurrentImaginaryOp>();
    builder_manager.addCreateFunction<BranchDataCurrentMagnitudeOp>();
    builder_manager.addCreateFunction<BranchDataCurrentPhaseOp>();
    builder_manager.addCreateFunction<BranchDataCurrentDecibelsOp>();
    builder_manager.addCreateFunction<BranchDataPosNegPowerOp>();
    builder_manager.addCreateFunction<BranchDataBJTPowerOp>();
    builder_manager.addCreateFunction<BranchDataJFETPowerOp>();
    builder_manager.addCreateFunction<BranchDataMESFETPowerOp>();
    builder_manager.addCreateFunction<BranchDataMOSFETPowerOp>();
    builder_manager.addCreateFunction<BranchDataTRAPowerOp>();
  }

  virtual Util::Op::Operator *makeOp(Util::ParamList::const_iterator &it) const {
    Util::Op::Operator *new_op = 0;
    const std::string &param_tag = (*it).tag();

    std::vector<std::string> args;
    std::string name;
    parameterNameAndArgs(name, args, it);

    if ((param_tag == "P") || (param_tag == "W"))
    {
      // This check may no longer be needed, now that P() and W() work in a .MEASURE statement.
      // Leaving it in for now.
      if (args.size() == 0)
      {
        Report::UserFatal0() << "Error building operator: " << name << " missing variable name";
      }
      // Node name could be circuit_context:DeviceTypeDeviceName while internally it should be DeviceType:circuit_context:DeviceName.
      std::string modifiedName = Util::xyceDeviceNameToSpiceName(args[0]);
            
      // could be a device lead current "DEV_I" or a branch current.
      // so we don't have to duplicate solution vars(branch currents) in the
      // store vector, look for each type.
      bool param_func = std::find(func_names, func_names + sizeof(func_names)/sizeof(func_names[0]), "I") != func_names + sizeof(func_names)/sizeof(func_names[0]);
      std::string store_name = modifiedName + ":BRANCH_D" ; // + (param_func ? "I" : "I");  // if it is in the state/store vec.
      // this if block allows for spaces in YPDE names as in I1(YPDE NAME)
      // we try to find devices based on store_name in the following blocks of code,
      // so do this modification now.
      std::string::size_type space = store_name.find_first_of(" ");
      if (space != std::string::npos)
      {
        if (space == 4 && store_name.substr(0, 4) == "YPDE")
        {
          store_name.replace(4, 1, ":");
        }
      }

      // Search store
      if( modifiedName[0] == 'Q' )
      {
        // BJT type.  
        // this should be encapsulated into the BJT class as some point
        NodeNameMap::const_iterator itB = outputManager_.getBranchVarsNodeMap().find(store_name+"B");
        NodeNameMap::const_iterator itE = outputManager_.getBranchVarsNodeMap().find(store_name+"E");
        NodeNameMap::const_iterator itC = outputManager_.getBranchVarsNodeMap().find(store_name+"C");
        NodeNameMap::const_iterator itS = outputManager_.getBranchVarsNodeMap().find(store_name+"S");
        if ( (itB != outputManager_.getBranchVarsNodeMap().end()) &&
             (itE != outputManager_.getBranchVarsNodeMap().end()) &&
             (itC != outputManager_.getBranchVarsNodeMap().end()))
		{
		  int indexB = (*itB).second;
		  int indexC = (*itC).second;
		  int indexE = (*itE).second;
		  int indexS = -1;
                  if (itS != outputManager_.getBranchVarsNodeMap().end())
                    indexS = (*itS).second;
		  
		  new_op = new BranchDataBJTPowerOp(name, indexB, indexC, indexE, indexS);
		}
      
      }
      else if( modifiedName[0] == 'J' )
      {
        // JFET type.  
        // this should be encapsulated into the JFET class as some point
        NodeNameMap::const_iterator itD = outputManager_.getBranchVarsNodeMap().find(store_name+"D");
        NodeNameMap::const_iterator itG = outputManager_.getBranchVarsNodeMap().find(store_name+"G");
        NodeNameMap::const_iterator itS = outputManager_.getBranchVarsNodeMap().find(store_name+"S");
        if ( (itD != outputManager_.getBranchVarsNodeMap().end()) &&
             (itG != outputManager_.getBranchVarsNodeMap().end()) &&
             (itS != outputManager_.getBranchVarsNodeMap().end()))
		{
		  int indexD = (*itD).second;
		  int indexG = (*itG).second;
		  int indexS = (*itS).second;
 		  new_op = new BranchDataJFETPowerOp(name, indexD, indexG, indexS);
		}
      
      }
      else if( modifiedName[0] == 'M' )
      {
        // MOSFET type.  
        // this should be encapsulated into the MOSFET class as some point
        NodeNameMap::const_iterator itD = outputManager_.getBranchVarsNodeMap().find(store_name+"D");
        NodeNameMap::const_iterator itG = outputManager_.getBranchVarsNodeMap().find(store_name+"G");
        NodeNameMap::const_iterator itS = outputManager_.getBranchVarsNodeMap().find(store_name+"S");
        NodeNameMap::const_iterator itB = outputManager_.getBranchVarsNodeMap().find(store_name+"B");
        if ( (itD != outputManager_.getBranchVarsNodeMap().end()) &&
             (itG != outputManager_.getBranchVarsNodeMap().end()) &&
             (itS != outputManager_.getBranchVarsNodeMap().end()))
	{
	  int indexD = (*itD).second;
	  int indexG = (*itG).second;
	  int indexS = (*itS).second;
	  int indexB = -1;
          if (itB != outputManager_.getBranchVarsNodeMap().end())
	  {
            indexB = (*itB).second;
	  }	  
	  new_op = new BranchDataMOSFETPowerOp(name, indexD, indexG, indexS, indexB);
        }
      }
      else if( modifiedName[0] == 'T')
      {
        // T Device (TRA or Lossless Transmission Line)
        NodeNameMap::const_iterator it1 = outputManager_.getBranchVarsNodeMap().find(store_name+"1");
        NodeNameMap::const_iterator it2 = outputManager_.getBranchVarsNodeMap().find(store_name+"2");
        if ( (it1 != outputManager_.getBranchVarsNodeMap().end()) &&
             (it2 != outputManager_.getBranchVarsNodeMap().end())) 
	{
	  int index1 = (*it1).second;
	  int index2 = (*it2).second;
 	  new_op = new BranchDataTRAPowerOp(name, index1, index2);
	}
      }
      else if( modifiedName[0] == 'Z' )
      {
        // MESFET type.  
        // this should be encapsulated into the MESFET class as some point
        NodeNameMap::const_iterator itD = outputManager_.getBranchVarsNodeMap().find(store_name+"D");
        NodeNameMap::const_iterator itG = outputManager_.getBranchVarsNodeMap().find(store_name+"G");
        NodeNameMap::const_iterator itS = outputManager_.getBranchVarsNodeMap().find(store_name+"S");
        if ( (itD != outputManager_.getBranchVarsNodeMap().end()) &&
             (itG != outputManager_.getBranchVarsNodeMap().end()) &&
             (itS != outputManager_.getBranchVarsNodeMap().end()))
	{
	  int indexD = (*itD).second;
	  int indexG = (*itG).second;
	  int indexS = (*itS).second;
 	  new_op = new BranchDataMESFETPowerOp(name, indexD, indexG, indexS);
	}
      }
      else
      {
        // default for devices for which power = I*V
	NodeNameMap::const_iterator it = outputManager_.getBranchVarsNodeMap().find(store_name);
	if (it != outputManager_.getBranchVarsNodeMap().end())
	{
	  int index = (*it).second;
          new_op = new BranchDataPosNegPowerOp(name, index);
        }
      }

      if (new_op)
      {
        new_op->addArg(args[0]);
      }
    }

    return new_op;
  }

private:
  const OutputMgr &     outputManager_;
};

//-------------------------------------------------------------------------- 
// Structure     : Util::Op::Builder::ExpressionOpBuilder 
// Purpose       : Makes an ExpressionOp or ConstantOp.
// Special Notes : 
// Creator       : Dave Baur
// Creation Date : 08/04/14 
//--------------------------------------------------------------------------
struct ExpressionOpBuilder : public Util::Op::Builder
{
  ExpressionOpBuilder(Parallel::Machine comm, const OutputMgr & output_manager)
    : comm_(comm),
      outputManager_(output_manager)
  {}

  virtual ~ExpressionOpBuilder()
  {}

  virtual void registerCreateFunctions(Util::Op::BuilderManager &builder_manager) const
  {
    builder_manager.addCreateFunction<ExpressionOp>();
    builder_manager.addCreateFunction<Util::Op::ConstantOp>();
  }

  virtual Util::Op::Operator *makeOp(Util::ParamList::const_iterator &it) const 
  {
    Util::Op::Operator *new_op = 0;
    const std::string &param_tag = (*it).tag();
    int param_type = (*it).getType();

    if (Util::hasExpressionTag(*it))
    {
      if (param_type == Util::EXPR)
      {
        Util::Param &param = const_cast<Util::Param &>(*it);

        new_op = new ExpressionOp(param_tag, param.getValue<Util::Expression>(), comm_, outputManager_);
      }
      // else if( (param_type == Util::DBLE) || (param_type == Util::INT) )
      // {
      //   new_op = new Util::Op::ConstantOp(param_tag, (*it).getImmutableValue<double>());
      // }
      else
      {
        new_op = new ExpressionOp(param_tag, param_tag, comm_, outputManager_);
      }
    }

    return new_op;
  }

private:
  const Parallel::Machine       comm_;
  const OutputMgr &             outputManager_;
};

//-------------------------------------------------------------------------- 
// Structure     : Util::Op::Builder::MeasurementOpBuilder 
// Purpose       : Builds a MeasurementOp.
// Special Notes : 
// Creator       : Dave Baur
// Creation Date : 08/04/14 
//--------------------------------------------------------------------------
struct MeasurementOpBuilder : public Util::Op::Builder
{
  MeasurementOpBuilder(const Measure::Manager & measure_manager)
    : measureManager_(measure_manager)
  {}

  virtual ~MeasurementOpBuilder()
  {}

  virtual void registerCreateFunctions(Util::Op::BuilderManager &builder_manager) const
  {
    builder_manager.addCreateFunction<MeasureOp>();
  }

  virtual Util::Op::Operator *makeOp(Util::ParamList::const_iterator &it) const 
  {
    Util::Op::Operator *new_op = 0;
    const std::string &param_tag = (*it).tag();

    const Measure::Base *measure = measureManager_.find(param_tag);
    if (measure)
    {
      new_op = new MeasureOp(param_tag, *measure);
    }

    return new_op;
  }

private:
  const Measure::Manager &             measureManager_;
};

//-------------------------------------------------------------------------- 
// Structure     : Util::Op::Builder::CircuitIndexOpBuilder 
// Purpose       : Builds a CircuitIndexOp.
// Special Notes : 
// Creator       : Dave Baur
// Creation Date : 08/04/14 
//--------------------------------------------------------------------------
struct CircuitIndexOpBuilder : public Util::Op::Builder
{
  CircuitIndexOpBuilder()
  {}

  virtual ~CircuitIndexOpBuilder()
  {}

  virtual void registerCreateFunctions(Util::Op::BuilderManager &builder_manager) const
  {
    builder_manager.addCreateFunction<CurrentIndexOp>();
  }

  virtual Util::Op::Operator *makeOp(Util::ParamList::const_iterator &it) const 
  {
    Util::Op::Operator *new_op = 0;
    const std::string &param_tag = (*it).tag();

    if (param_tag == "INDEX") 
    {
      new_op  = new CurrentIndexOp(param_tag);
    }

    return new_op;
  }
};

//-------------------------------------------------------------------------- 
// Structure     : Util::Op::Builder::SensitivityOpBuilder 
// Purpose       : Builds various "sensitivity operators".  See the
//                 list below in registerCreateFunctions(). This includes 
//                 the operators for both direct and adjoint sensitivities.
// Special Notes : 
// Creator       : Dave Baur
// Creation Date : 08/04/14 
//--------------------------------------------------------------------------
struct SensitivityOpBuilder : public Util::Op::Builder
{
  SensitivityOpBuilder()
  {}

  virtual ~SensitivityOpBuilder()
  {}

  virtual void registerCreateFunctions(Util::Op::BuilderManager &builder_manager) const
  {
    builder_manager.addCreateFunction<SensitivityObjFunctionOp>();
    builder_manager.addCreateFunction<SensitivitydOdpDirectOp>();
    builder_manager.addCreateFunction<SensitivitydOdpDirectScaledOp>();
    builder_manager.addCreateFunction<SensitivitydOdpAdjointOp>();
    builder_manager.addCreateFunction<SensitivitydOdpAdjointScaledOp>();
  }

  virtual Util::Op::Operator *makeOp(Util::ParamList::const_iterator &it) const 
  {
    Util::Op::Operator *new_op = 0;
    const std::string &param_tag = (*it).tag();
    const std::string &param_string = (*it).stringValue();

    // Certain variables were added to this list with the temporary tag "SENS",
    // in the OutputMgr::registerSens function.  They also contain 
    // OP information, which can be used here to construct the proper name
    // for each derivative.
    if (param_tag == "SENS")
    {
      std::string function;
      std::string parameter;
      ptrdiff_t type;
      int index;

      Util::Marshal min(param_string);

      min >> function >> parameter >> type >> index;

      std::string name = "d" + function + "/d(" + parameter + ")";
      if (type == Util::Op::identifier<SensitivityObjFunctionOp>())
      {
        new_op = new SensitivityObjFunctionOp(function, index);
      }
      else if (type == Util::Op::identifier<SensitivitydOdpDirectOp>())
      {
        new_op = new SensitivitydOdpDirectOp(name + "_Dir", index);
      }
      else if (type == Util::Op::identifier<SensitivitydOdpDirectScaledOp>())
      {
        new_op = new SensitivitydOdpDirectScaledOp(name + "_Dir_scaled", index);
      }
      else if (type == Util::Op::identifier<SensitivitydOdpAdjointOp>())
      {
        new_op = new SensitivitydOdpAdjointOp(name + "_Adj", index);
      }
      if (type == Util::Op::identifier<SensitivitydOdpAdjointScaledOp>())
      {
        new_op = new SensitivitydOdpAdjointScaledOp(name + "_Adj_scaled", index);
      }
    }

    return new_op;
  }
};

//-------------------------------------------------------------------------- 
// Structure     : Util::Op::Builder::TransientAdjointOpBuilder 
// Purpose       : Builder for various "sensitivity operators".  See
//                 the list below in registerCreateFunctions().
// Special Notes : 
// Creator       : Dave Baur
// Creation Date : 08/04/14 
//--------------------------------------------------------------------------
struct TransientAdjointOpBuilder : public Util::Op::Builder
{
  TransientAdjointOpBuilder()
  {}

  virtual ~TransientAdjointOpBuilder()
  {}

  virtual void registerCreateFunctions(Util::Op::BuilderManager &builder_manager) const
  {
    builder_manager.addCreateFunction<SensitivityObjFunctionOp>();
    builder_manager.addCreateFunction<SensitivitydOdpDirectOp>();
    builder_manager.addCreateFunction<SensitivitydOdpDirectScaledOp>();
    builder_manager.addCreateFunction<SensitivitydOdpAdjointOp>();
    builder_manager.addCreateFunction<SensitivitydOdpAdjointScaledOp>();
  }

  virtual Util::Op::Operator *makeOp(Util::ParamList::const_iterator &it) const 
  {
    Util::Op::Operator *new_op = 0;
    const std::string &param_tag = (*it).tag();
    const std::string &param_string = (*it).stringValue();

    // Certain variables were added to this list with the temporary tag "SENS",
    // in the OutputMgr::registerSens function.  They also contain 
    // OP information, which can be used here to construct the proper name
    // for each derivative.
    if (param_tag == "SENS")
    {
      std::string function;
      std::string parameter;
      ptrdiff_t type;
      int index;

      Util::Marshal min(param_string);

      min >> function >> parameter >> type >> index;

      std::string name = "d" + function + "/d(" + parameter + ")";
      if (type == Util::Op::identifier<SensitivityObjFunctionOp>())
      {
        new_op = new SensitivityObjFunctionOp(function, index);
      }
      else if (type == Util::Op::identifier<SensitivitydOdpDirectOp>())
      {
        new_op = new SensitivitydOdpDirectOp(name + "_Dir", index);
      }
      else if (type == Util::Op::identifier<SensitivitydOdpDirectScaledOp>())
      {
        new_op = new SensitivitydOdpDirectScaledOp(name + "_Dir_scaled", index);
      }
      else if (type == Util::Op::identifier<SensitivitydOdpAdjointOp>())
      {
        new_op = new SensitivitydOdpAdjointOp(name + "_Adj", index);
      }
      if (type == Util::Op::identifier<SensitivitydOdpAdjointScaledOp>())
      {
        new_op = new SensitivitydOdpAdjointScaledOp(name + "_Adj_scaled", index);
      }
    }

    return new_op;
  }
};

//-------------------------------------------------------------------------- 
// Function      : registerOpBuilders 
// Purpose       : register various builders with the op_builder_manager
// Special Notes : 
// Creator       : Dave Baur
// Creation Date : 08/04/14 
//--------------------------------------------------------------------------
void registerOpBuilders(Util::Op::BuilderManager &op_builder_manager, Parallel::Machine comm, OutputMgr &output_manager)
{
  op_builder_manager.addBuilder(new CircuitTemperatureOpBuilder(output_manager));
  op_builder_manager.addBuilder(new CircuitTimeOpBuilder(output_manager));
  op_builder_manager.addBuilder(new CircuitNoiseContOpBuilder(output_manager));
  op_builder_manager.addBuilder(new CircuitOutputNoiseOpBuilder(output_manager));
  op_builder_manager.addBuilder(new CircuitInputNoiseOpBuilder(output_manager));
  op_builder_manager.addBuilder(new CircuitFrequencyOpBuilder(output_manager));
  op_builder_manager.addBuilder(new CircuitIndexOpBuilder());
  op_builder_manager.addBuilder(new SensitivityOpBuilder());
  op_builder_manager.addBuilder(new TransientAdjointOpBuilder());
  op_builder_manager.addBuilder(new ExpressionOpBuilder(comm, output_manager));
//  op_builder_manager.addBuilder(new StepSweepOpBuilder(output_manager));
  op_builder_manager.addBuilder(new DCSweepOpBuilder(output_manager));
  op_builder_manager.addBuilder(new DCSweepCurrentValueOpBuilder(output_manager));
  op_builder_manager.addBuilder(new InternalVariableOpBuilder(output_manager));
  op_builder_manager.addBuilder(new VoltageVariableOpBuilder(output_manager));
  op_builder_manager.addBuilder(new CurrentVariableOpBuilder(output_manager));
  op_builder_manager.addBuilder(new PowerVariableOpBuilder(output_manager));
}

//-------------------------------------------------------------------------- 
// Function      : registerOpBuilders 
// Purpose       : register MeasurementOpBuilder with the op_builder_manager
// Special Notes : 
// Creator       : Dave Baur
// Creation Date : 08/04/14 
//--------------------------------------------------------------------------
void registerOpBuilders(Util::Op::BuilderManager &op_builder_manager, Parallel::Machine comm, Measure::Manager &measure_manager)
{
  op_builder_manager.addBuilder(new MeasurementOpBuilder(measure_manager));
}

} // namespace IO
} // namespace Xyce
