//-------------------------------------------------------------------------
//   Copyright 2002-2018 National Technology & Engineering Solutions of
//   Sandia, LLC (NTESS).  Under the terms of Contract DE-NA0003525 with
//   NTESS, the U.S. Government retains certain rights in this software.
//
//   This file is part of the Xyce(TM) Parallel Electrical Simulator.
//
//   Xyce(TM) is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   Xyce(TM) is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with Xyce(TM).
//   If not, see <http://www.gnu.org/licenses/>.
//-------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
// Purpose        : interface to preconditioner
//
// Special Notes  :
//
// Creator        : Heidi Thornquist, SNL, Electrical & Microsystem Modeling
//
// Creation Date  : 09/27/07
//
//
//
//
//-----------------------------------------------------------------------------

#ifndef Xyce_N_LAS_MLPrecond_h
#define Xyce_N_LAS_MLPrecond_h

// ---------- Standard Includes ----------

// ----------   Xyce Includes   ----------

#include <N_LAS_Preconditioner.h>
#include <N_LAS_Problem.h>
#include <N_UTL_OptionBlock.h>

#include <Teuchos_ParameterList.hpp>

class Epetra_Operator;
namespace ML_Epetra {
  class MultiLevelPreconditioner;
}

namespace Xyce {
namespace Linear {

//-----------------------------------------------------------------------------
// Class         : N_LAS_MLPrecond
// Purpose       : interface to ifpack preconditioner
// Special Notes :
// Creator       : Heidi Thornquist, SNL, Electrical & Microsystem Modeling
// Creation Date : 09/27/07
//-----------------------------------------------------------------------------
class N_LAS_MLPrecond : public N_LAS_Preconditioner
{

public:
  // Constructors
  N_LAS_MLPrecond();

  // Destructor
  virtual ~N_LAS_MLPrecond() {}

  // Set the preconditioner options
  bool setOptions(const Util::OptionBlock & OB);
  bool setDefaultOptions();
  bool setDefaultOption( const string & option );

  // Set individual preconditioner options
  bool setParam( const Util::Param & param );

  // Set or reset the matrix pattern for the preconditioner using problem.
  // \note The preconditioner will be recreated each time this is called.
  bool initGraph( const Teuchos::RCP<N_LAS_Problem> & problem );

  // Set the matrix values for the preconditioner
  bool initValues( const Teuchos::RCP<N_LAS_Problem> & problem );

  // Compute the preconditioner using the current matrix values.
  bool compute();

  // Apply the preconditioner; y = M*x.
  int apply( N_LAS_MultiVector & x, N_LAS_MultiVector & y );

  // Return the preconditioner as an Epetra_Operator object.
  Teuchos::RCP<Epetra_Operator> epetraObj() { return epetraPrec_; }

private:
  // Maximum number of levels.
  int maxLevel_;

  // ML_Epetra preconditioning class.
  Teuchos::RCP<ML_Epetra::MultiLevelPreconditioner> mlPrecond_;

  // Parameter list containing preconditioning options
  Teuchos::ParameterList mlList_;

  // Current matrix being preconditioned.
  Teuchos::RCP<N_LAS_Problem> problem_;

  // Preconditioner as an Epetra_Operator object.
  Teuchos::RCP<Epetra_Operator> epetraPrec_;

  // Options block.
  Teuchos::RCP<N_UTL_OptionBlock> options_;

  // Default preconditioner values
  static const int maxLevel_default_ = 5;

  // No copying
  N_LAS_MLPrecond(const N_LAS_MLPrecond & right);
  N_LAS_MLPrecond & operator=(const N_LAS_MLPrecond & right);

  // No comparison
  bool operator==(const N_LAS_MLPrecond & right) const;
  bool operator!=(const N_LAS_MLPrecond & right) const;

};

} // namespace Linear
} // namespace Xyce

#endif
