//-------------------------------------------------------------------------
//   Copyright 2002-2018 National Technology & Engineering Solutions of
//   Sandia, LLC (NTESS).  Under the terms of Contract DE-NA0003525 with
//   NTESS, the U.S. Government retains certain rights in this software.
//
//   This file is part of the Xyce(TM) Parallel Electrical Simulator.
//
//   Xyce(TM) is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   Xyce(TM) is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with Xyce(TM).
//   If not, see <http://www.gnu.org/licenses/>.
//-------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
// Purpose        : C callable functions to link to Xyce's C++ interface
//
// Special Notes  : These functions expose Xyce's C++ interface to 
//                  other programs (i.e. Python) that can dynamically 
//                  link to a C library (via ctypes in Python)
//
//
//-----------------------------------------------------------------------------


// ---------- Standard Includes ----------
#include <iostream>
#include <ctime>

// ----------   Xyce Includes   ----------
#include <Xyce_config.h>
#include <N_CIR_Xyce.h>
#include <N_DEV_Device.h>
#include <N_DEV_Algorithm.h>
#include <N_DEV_ADC.h>

#include <N_CIR_XyceCInterface.h>

void xyce_open( void ** ptr)
{
  N_CIR_Xyce * xycePtr = new N_CIR_Xyce();
  *ptr = xycePtr;
}


void xyce_close( void** ptr)
{
  N_CIR_Xyce * xycePtr = static_cast<N_CIR_Xyce *>( *ptr );
  xycePtr->finalize();
  delete xycePtr;
  std::cout << "xyce_close after delete xycePtr\n" ;
  *ptr = 0;
}

int xyce_initialize( void** ptr, int argc, char ** argv )
{
  N_CIR_Xyce * xycePtr = static_cast<N_CIR_Xyce *>( *ptr );
  return( xycePtr->initialize( argc, argv ) );
}


int xyce_runSimulation(void ** ptr)
{
  N_CIR_Xyce * xycePtr = static_cast<N_CIR_Xyce *>( *ptr );
  return(xycePtr->runSimulation());
}

int xyce_simulateUntil(  void **ptr, double requestedUntilTime, double & completedUntilTime )
{
  N_CIR_Xyce * xycePtr = static_cast<N_CIR_Xyce *>( *ptr );
  bool retVal = xycePtr->simulateUntil( requestedUntilTime,  completedUntilTime);
  if (retVal)
    return 1;
  return 0;
}


int xyce_getDeviceNames(void ** ptr, char * modelGroupName, int & numDevNames, char ** deviceNames)
{
  N_CIR_Xyce * xycePtr = static_cast<N_CIR_Xyce *>( *ptr );
  const std::string modelGroupNameString( modelGroupName );
  std::vector<std::string> deviceNamesVec;
  xycePtr->getDeviceNames(modelGroupNameString , deviceNamesVec);
  numDevNames = deviceNamesVec.size();
  for( int i=0;i < deviceNamesVec.size(); i++ )
  {
    std::strcpy( (deviceNames)[i], deviceNamesVec.at(i).c_str() );
  }
  return 0;
}


int xyce_getDACDeviceNames(void ** ptr, int & numDevNames, char ** deviceNames)
{
  N_CIR_Xyce * xycePtr = static_cast<N_CIR_Xyce *>( *ptr );
  std::vector<std::string> deviceNamesVec;
  xycePtr->getDACDeviceNames( deviceNamesVec );
  numDevNames = deviceNamesVec.size();
  for( int i=0;i < deviceNamesVec.size(); i++ )
  {
    std::strcpy( (deviceNames)[i], deviceNamesVec.at(i).c_str() );
  }
  return 0;
}


int xyce_getADCMap(void ** ptr) /* std::map<std::string,std::map<std::string,double> >& ADCMap); */
{
  N_CIR_Xyce * xycePtr = static_cast<N_CIR_Xyce *>( *ptr );
  return 0;
}

int xyce_updateTimeVoltagePairs(void ** ptr, char * DACname, int numPoints, double * timeArray, double * voltageArray)
{
  N_CIR_Xyce * xycePtr = static_cast<N_CIR_Xyce *>( *ptr );

  std::string name(DACname);
  std::map< std::string, std::vector<std::pair<double,double> > *> timePointsMap;
  std::vector<std::pair<double,double> > storageVec;
  if( numPoints > 0) {
    for( int i=0 ; i<numPoints; i++ ) {
      storageVec.push_back( std::pair<double,double>( timeArray[i], voltageArray[i] ) );
    }
  }
  timePointsMap[ name ] = &storageVec;
  /*
  std::cout << "name= \"" << name << "\"" << std::endl;
  for( int i=0; i<numPoints; i++ )
    std::cout << "storageVec[ " << i << "] = " << storageVec[i].first << ", " << storageVec[i].second << std::endl;
  */
  xycePtr->updateTimeVoltagePairs(timePointsMap);
  
  return 0;
}


int xyce_getTimeVoltagePairsADC( void** ptr, int * numADCnames, char ** ADCnames, int * numPoints, double ** timeArray, double ** voltageArray )
{
  N_CIR_Xyce * xycePtr = static_cast<N_CIR_Xyce *>( *ptr );
 
  std::map< std::string, std::vector< std::pair<double,double> > > timeVoltageUpdateMap;
  int status = xycePtr->getTimeVoltagePairs( timeVoltageUpdateMap );
  
  std::map< std::string, std::vector< std::pair<double,double> > >::iterator cMap = timeVoltageUpdateMap.begin();
  std::map< std::string, std::vector< std::pair<double,double> > >::iterator eMap = timeVoltageUpdateMap.end();
  int ADCnum = 0;
  while( cMap != eMap )
  {
    strcpy( ADCnames[ADCnum], (cMap->first).c_str() );
    std::vector< std::pair<double,double> > dataVec = cMap->second;
    for( int j = 0; j<dataVec.size(); j++ )
    {
      timeArray[ADCnum][j] = dataVec[j].first;
      voltageArray[ADCnum][j] = dataVec[j].second;
    }
    ADCnum++;
    ++cMap;
  }
  *numADCnames=ADCnum;
  
  return status;
}

int xyce_setADCWidths(void ** ptr, int numADCnames, char ** ADCnames, int *  widths) 
{
  N_CIR_Xyce * xycePtr = static_cast<N_CIR_Xyce *>( *ptr );
  
  std::map<std::string, int> ADCWidthMap;
  for( int i=0; i<numADCnames; i++ )
  {
    ADCWidthMap[ std::string(ADCnames[i]) ] = widths[i];
  }
  
  int status = xycePtr->setADCWidths( ADCWidthMap);
  return status;
}

int xyce_checkResponseVar(void ** ptr, char * variable_name)
{
  N_CIR_Xyce * xycePtr = static_cast<N_CIR_Xyce *>( *ptr );
  std::string variable_nameStr( variable_name );
  int result = xycePtr->checkResponseVar(  variable_nameStr );
  return result;
}

int xyce_obtainResponse(void ** ptr, char * variable_name, double &value) 
{
  N_CIR_Xyce * xycePtr = static_cast<N_CIR_Xyce *>( *ptr );
  std::string variable_nameStr( variable_name );
  int result = xycePtr->obtainResponse(  variable_nameStr,  value  );
  return result;
}








